from decimal import Decimal
from viggofiscal.subsystem.calculo_fiscal.icms.base_reduzida_icms_proprio \
    import BaseReduzidaIcmsProprio
from viggofiscal.subsystem.calculo_fiscal.icms.base_icms_proprio \
    import BaseIcmsProprio
from viggofiscal.subsystem.calculo_fiscal.icms.icms00 \
    import Icms00
from viggofiscal.subsystem.calculo_fiscal.icms.valor_icms_proprio \
    import ValorIcmsProprio
from viggofiscal.subsystem.calculo_fiscal.utils import round_abnt, to_decimal
from viggofiscal.subsystem.calculo_fiscal.icms.base_icms_st \
    import BaseIcmsST
from viggofiscal.subsystem.calculo_fiscal.icms.base_reduzida_icms_st \
    import BaseReduzidaIcmsST
from viggofiscal.subsystem.calculo_fiscal.icms.valor_icms_st \
    import ValorIcmsST


class Icms90():

    def __init__(self, valor_produto: Decimal, valor_frete: Decimal,
                 valor_seguro: Decimal, despesas_acessorias: Decimal,
                 valor_desconto: Decimal,
                 aliq_icms_proprio: Decimal, aliq_icms_st: Decimal,
                 mva: Decimal, valor_ipi: Decimal=Decimal('0.0'),
                 percentual_reducao: Decimal=Decimal('0.0'),
                 percentual_reducao_st: Decimal=Decimal('0.0')):
        self.valor_produto = to_decimal(valor_produto)
        self.valor_frete = to_decimal(valor_frete)
        self.valor_seguro = to_decimal(valor_seguro)
        self.despesas_acessorias = to_decimal(despesas_acessorias)
        self.valor_desconto = to_decimal(valor_desconto)
        self.aliq_icms_proprio = to_decimal(aliq_icms_proprio)
        self.aliq_icms_st = to_decimal(aliq_icms_st)
        self.mva = to_decimal(mva)
        self.valor_ipi = to_decimal(valor_ipi)
        self.percentual_reducao = to_decimal(percentual_reducao)
        self.percentual_reducao_st = to_decimal(percentual_reducao_st)

    # ICMS Próprio
    def base_icms_proprio(self) -> Decimal:
        self.bc_icms_proprio = BaseIcmsProprio(
            self.valor_produto, self.valor_frete, self.valor_seguro,
            self.despesas_acessorias, self.valor_desconto, self.valor_ipi).\
            calcular_base_icms_proprio()
        return self.bc_icms_proprio

    def base_reduzida_icms_proprio(self) -> Decimal:
        self.bc_reduzida_icms_proprio = BaseReduzidaIcmsProprio(
            self.valor_produto, self.valor_frete, self.valor_seguro,
            self.despesas_acessorias, self.valor_desconto,
            self.percentual_reducao, self.valor_ipi).\
            calcular_base_reduzida_icms_proprio()
        return self.bc_reduzida_icms_proprio

    def valor_icms_proprio(self) -> Decimal:
        valor_icms_proprio = ValorIcmsProprio(
            self.base_icms_proprio(), self.aliq_icms_proprio).\
            calcular_valor_icms_proprio()
        return valor_icms_proprio

    def valor_icms_proprio_base_reduzida(self) -> Decimal:
        valor_icms_proprio = ValorIcmsProprio(
            self.base_reduzida_icms_proprio(), self.aliq_icms_proprio).\
            calcular_valor_icms_proprio()
        return valor_icms_proprio

    def valor_icms_proprio_desonerado(self) -> Decimal:
        icms00 = Icms00(
            self.valor_produto, self.valor_frete, self.valor_seguro,
            self.despesas_acessorias, self.valor_ipi, self.valor_desconto,
            self.aliq_icms_proprio)
        valor_icms_normal = icms00.valor_icms_proprio()
        valor_icms_desonerado = (
            valor_icms_normal - self.valor_icms_proprio_base_reduzida())
        return round_abnt(valor_icms_desonerado, 2)

    # ICMS ST
    def base_icms_st(self) -> Decimal:
        self.bc_icms_st = BaseIcmsST(self.base_icms_proprio(), self.mva,
                                     self.valor_ipi)
        return self.base_icms_st().calcular_base_icms_st()

    def base_reduzida_icms_st(self) -> Decimal:
        self.bc_reduzida_icms_st = BaseReduzidaIcmsST(
            self.base_icms_proprio(), self.mva, self.percentual_reducao_st,
            self.valor_ipi)
        return self.bc_reduzida_icms_st.calcular_base_reduzida_icms_st()

    def valor_icms_st(self) -> Decimal:
        valor_icms_st = ValorIcmsST(
            self.base_icms_st(), self.aliq_icms_st, self.valor_icms_proprio()).\
            calcular_valor_icms_st()
        return valor_icms_st
    
    def valor_icms_st_base_reuzida(self) -> Decimal:
        valor_icms_st_base_reduzida = ValorIcmsST(
            self.base_reduzida_icms_st(), self.aliq_icms_st,
            self.valor_icms_proprio()).calcular_valor_icms_st()
        return valor_icms_st_base_reduzida

    def valor_icms_st_desonerado(self) -> Decimal:
        icms10 = Icms10(
            self.valor_produto, self.valor_frete, self.valor_seguro,
            self.despesas_acessorias, self.valor_ipi, self.valor_desconto,
            self.aliq_icms_proprio, self.aliq_icms_st, self.mva)
        valor_icms_st_normal = icms10.valor_icms_st()
        valor_icms_st_desonerado = (
            valor_icms_st_normal - self.valor_icms_st_base_reuzida())

        return round_abnt(valor_icms_st_desonerado, 2)



from sqlalchemy import orm
from viggocore.database import db
from viggocore.common.subsystem import entity
from viggocore.common import exception


class RegraFiscal(entity.Entity, db.Model):

    attributes = ['domain_org_id', 'cstcofins_id', 'csticms_id', 'cstpis_id',
                  'cfop_id', 'cstipi_id', 'descricao', 'aliq_especifica_icms',
                  'aliq_especifica_icms_st', 'aliq_reducao_base_icms',
                  'aliq_reducao_base_st', 'aliq_diferimento', 'indice_difal',
                  'difal_base_dupla', 'icms_sob_pis_cofins', 'icms_sob_frete',
                  'tipo_ipi', 'icms_sob_ipi', 'codigo_cst_icms']
    attributes += entity.Entity.attributes

    domain_org_id = db.Column(
        db.CHAR(32), db.ForeignKey('domain_org.id'), nullable=False)
    cstcofins_id = db.Column(
        db.CHAR(32), db.ForeignKey('cstcofins.id'), nullable=False)
    cstcofins = orm.relationship('Cstcofins', backref=orm.backref('cstcofinss'))
    csticms_id = db.Column(
        db.CHAR(32), db.ForeignKey('csticms.id'), nullable=False)
    csticms = orm.relationship('Csticms', backref=orm.backref('csticmss'))
    cstpis_id = db.Column(
        db.CHAR(32), db.ForeignKey('cstpis.id'), nullable=False)
    cstpis = orm.relationship('Cstpis', backref=orm.backref('cstpiss'))
    cfop_id = db.Column(
        db.CHAR(32), db.ForeignKey('cfop.id'), nullable=False)
    cfop = orm.relationship('Cfop', backref=orm.backref('cfops'))
    cstipi_id = db.Column(
        db.CHAR(32), db.ForeignKey('cstipi.id'), nullable=True)
    cstipi = orm.relationship('Cstipi', backref=orm.backref('cstipis'))

    descricao = db.Column(db.String(100), nullable=False)
    aliq_especifica_icms = db.Column(
        db.Numeric(5, 2), nullable=False, default=0, server_default="0")
    aliq_especifica_icms_st = db.Column(
        db.Numeric(5, 2), nullable=False, default=0, server_default="0")
    aliq_reducao_base_icms = db.Column(
        db.Numeric(5, 2), nullable=False, default=0, server_default="0")
    aliq_reducao_base_st = db.Column(
        db.Numeric(5, 2), nullable=False, default=0, server_default="0")
    aliq_diferimento = db.Column(
        db.Numeric(5, 2), nullable=False, default=0, server_default="0")
    indice_difal = db.Column(db.Boolean, nullable=False)
    difal_base_dupla = db.Column(db.Boolean, nullable=False)
    icms_sob_pis_cofins = db.Column(db.Boolean, nullable=False)
    icms_sob_frete = db.Column(db.Boolean, nullable=False)
    tipo_ipi = db.Column(db.CHAR(6), nullable=True)
    icms_sob_ipi = db.Column(db.Boolean, nullable=True)

    def __init__(self, id, domain_org_id, cstcofins_id, csticms_id,
                 cfop_id, cstipi_id, descricao, indice_difal, difal_base_dupla,
                 icms_sob_pis_cofins, icms_sob_frete,
                 aliq_especifica_icms=0, aliq_especifica_icms_st=0,
                 aliq_reducao_base_icms=0, aliq_reducao_base_st=0,
                 aliq_diferimento=0, cstpis_id=None, tipo_ipi=None,
                 icms_sob_ipi=None,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.domain_org_id = domain_org_id
        self.cstcofins_id = cstcofins_id
        self.csticms_id = csticms_id
        self.cfop_id = cfop_id
        self.cstipi_id = cstipi_id
        self.descricao = descricao
        self.aliq_especifica_icms = aliq_especifica_icms
        self.aliq_especifica_icms_st = aliq_especifica_icms_st
        self.aliq_reducao_base_icms = aliq_reducao_base_icms
        self.aliq_reducao_base_st = aliq_reducao_base_st
        self.indice_difal = indice_difal
        self.difal_base_dupla = difal_base_dupla
        self.icms_sob_pis_cofins = icms_sob_pis_cofins
        self.icms_sob_frete = icms_sob_frete
        self.cstpis_id = cstpis_id
        self.tipo_ipi = tipo_ipi
        self.icms_sob_ipi = icms_sob_ipi
        self.aliq_diferimento = aliq_diferimento

    def is_stable(self):
        # validação se icms_sob_ipi
        if (self.icms_sob_ipi is True and
            (self.tipo_ipi is None and self.cstipi_id is None)):
            raise exception.BadRequest('Se icms_sob_ipi é verdadeiro então' +
                'tipo_ipi e cstipi_id devem ser informados.')

        # TODO(JorgeSilva): verificar como fazer essas validações do tipo de
        # icms pois na criação eu ainda n tenho os backrefs.
        # # validações por tipo de icms
        # if self.csticms.cstcsosn == 20 and self.aliq_reducao_base_icms <= 0:
        #     raise exception.BadRequest('Se o cst_icms for igual a 20 a ' +
        #         'aliq_reducao_base_icms tem que ser maior que zero.')
        # elif self.csticms.cstcsosn == 51 and self.aliq_diferimento <= 0:
        #     raise exception.BadRequest('Se o cst_icms for igual a 51 a ' +
        #         'aliq_diferimento tem que ser maior que zero.')
        return True

    @property
    def codigo_cst_icms(self):
        return self.csticms.cstcsosn

    @classmethod
    def individual(cls):
        return 'regra_fiscal'



from decimal import Decimal


def to_decimal(num):
    try:
        return Decimal(str(num))
    except Exception:
        return Decimal('0.0')


def to_decimal_n(num, n):
    try:
        num_str = '{:.' + str(n) + 'f}'
        num_str = num_str.format(num)
        return Decimal(num_str)
    except Exception:
        return Decimal('0.0')


def is_decimal(num):
    response = False
    if (to_decimal(num) - int(num)) > to_decimal('0.0'):
        response = True
    return response


def normalize_number_to_print(valor):
    valor_decimal = to_decimal(valor)
    if valor_decimal - int(valor_decimal) != 0:
        return '{:.2f}'.format(valor_decimal)
    else:
        return int(valor_decimal)


# verifica se um número é par
def se_e_par(n):
    n = int(n)
    if (n % 2) == 0:
        return True
    else:
        return False


# função criada para arredondar um valor usando a norma ABNT 5891/77
# onde valor=o valor a ser arredondado
# n=limitador da quantidade de casas
def round_abnt(valor, n: int):
    try:
        vl_str = str(valor)
        if '.' in vl_str:
            inteira, decimal = str(valor).split('.')
        else:
            inteira = str(valor)
            decimal = '00'
    except Exception:
        print('Erro na conversão do valor!')

    decimal = (decimal + ('0' * 10))
    if len(decimal) > (n+2):
        prox = int(decimal[n])
        pos_prox = int(decimal[n+1])
        if ((prox >= 5 and pos_prox != 0) or
           (prox == 5 and pos_prox == 0 and se_e_par(decimal[n-1]) is False)):
            aux = int(decimal[n-1]) + 1
            decimal = decimal[:n-1] + str(aux) + decimal[n:]
    decimal = decimal[:n]
    return Decimal(f'{inteira}.{decimal}')



from decimal import Decimal
from viggofiscal.subsystem.calculo_fiscal.icms.base_reduzida_icms_proprio \
    import BaseReduzidaIcmsProprio
from viggofiscal.subsystem.calculo_fiscal.icms.base_icms_proprio \
    import BaseIcmsProprio
from viggofiscal.subsystem.calculo_fiscal.icms.icms00 \
    import Icms00
from viggofiscal.subsystem.calculo_fiscal.icms.valor_icms_proprio \
    import ValorIcmsProprio
from viggofiscal.subsystem.calculo_fiscal.utils import round_abnt, to_decimal
from viggofiscal.subsystem.calculo_fiscal.icms.base_icms_st \
    import BaseIcmsST
from viggofiscal.subsystem.calculo_fiscal.icms.base_reduzida_icms_st \
    import BaseReduzidaIcmsST
from viggofiscal.subsystem.calculo_fiscal.icms.valor_icms_st \
    import ValorIcmsST


class Icms101():

    def __init__(self, valor_produto: Decimal, valor_frete: Decimal,
                 valor_seguro: Decimal, despesas_acessorias: Decimal,
                 valor_desconto: Decimal, percentual_credito_sn: Decimal,
                 percentual_reducao: Decimal=Decimal('0.0')):
        self.valor_produto = to_decimal(valor_produto)
        self.valor_frete = to_decimal(valor_frete)
        self.valor_seguro = to_decimal(valor_seguro)
        self.despesas_acessorias = to_decimal(despesas_acessorias)
        self.valor_desconto = to_decimal(valor_desconto)
        self.percentual_credito_sn = to_decimal(percentual_credito_sn)
        self.percentual_reducao = to_decimal(percentual_reducao)

    def base_icms_proprio(self):
        if self.percentual_reducao == 0:
            self.bc_icms_proprio = BaseIcmsProprio(
                self.valor_produto, self.valor_frete, self.valor_seguro,
                self.despesas_acessorias, self.valor_desconto)
            return self.bc_icms_proprio.calcular_base_icms_proprio()
        else:
            self.bc_reduzida_icms_proprio = BaseReduzidaIcmsProprio(
                self.valor_produto, self.valor_frete, self.valor_seguro,
                self.despesas_acessorias, self.valor_desconto,
                self.percentual_reducao)
            return self.bc_reduzida_icms_proprio.\
                calcular_base_reduzida_icms_proprio()

    def valor_credito_sn(self):
        valor_credito_sn = (
            self.base_icms_proprio() * (self.percentual_credito_sn / 100))
        return round_abnt(valor_credito_sn, 2)



from decimal import Decimal
from viggofiscal.subsystem.calculo_fiscal.utils import round_abnt, to_decimal


class BaseIcmsProprio():

    def __init__(self, valor_produto: Decimal, valor_frete: Decimal,
                 valor_seguro: Decimal, despesas_acessorias: Decimal,
                 valor_desconto: Decimal, valor_ipi: Decimal=Decimal('0.0')):
        self.valor_produto = to_decimal(valor_produto)
        self.valor_frete = to_decimal(valor_frete)
        self.valor_seguro = to_decimal(valor_seguro)
        self.despesas_acessorias = to_decimal(despesas_acessorias)
        self.valor_desconto = to_decimal(valor_desconto)
        self.valor_ipi = to_decimal(valor_ipi)

    def calcular_base_icms_proprio(self) -> Decimal:
        base_icms_proprio = (
            self.valor_produto + self.valor_frete + self.valor_seguro +
            self.despesas_acessorias + self.valor_ipi - self.valor_desconto)
        return round_abnt(base_icms_proprio, 2)



from decimal import Decimal
from viggofiscal.subsystem.calculo_fiscal.utils import round_abnt, to_decimal


class BaseIcmsST():

    def __init__(self, base_icms_proprio: Decimal, mva: Decimal,
                 valor_ipi: Decimal=Decimal('0.0')):
        self.base_icms_proprio = to_decimal(base_icms_proprio)
        self.mva = to_decimal(mva)
        self.valor_ipi = to_decimal(valor_ipi)

    def calcular_base_icms_st(self) -> Decimal:
        base_icms_st = (
            (self.base_icms_proprio + self.valor_ipi) * (1 + (self.mva / 100)))
        return round_abnt(base_icms_st, 2)
